# -*- coding: utf-8 -*-
from .Base import Base
from .Database import db
from .Fields import Fields
from .Types import Types
from .Users import Users

__all__ = ['db', 'Base', 'Fields', 'Types', 'Users']



# -*- coding: utf-8 -*-
import os

from efesto.models import Fields, Types

from ruamel.yaml import YAML


class Blueprints:

    def __init__(self):
        self.yaml = YAML(typ='safe')

    @staticmethod
    def make_field(field_name, type_id, **options):
        field = Fields.create(name=field_name, type_id=type_id, owner_id=1,
                              **options)
        field.save()

    @staticmethod
    def options(options):
        dictionary = {}
        for option in options:
            option_name = list(option.keys())[0]
            dictionary[option_name] = option[option_name]
        return dictionary

    def load_field(self, new_type, field):
        """
        Loads a field in the database. If a field section is specified, parse
        it.
        """
        if isinstance(field, str):
            return self.make_field(field, new_type.id)
        for field_name, options in field.items():
            options_dict = self.options(options)
            return self.make_field(field_name, new_type.id, **options_dict)

    @staticmethod
    def load_type(table):
        """
        Loads a type in the database
        """
        return Types.create(name=table, owner_id=1)

    def read(self, blueprint):
        """
        Finds and reads blueprint
        """
        path = os.path.join(os.getcwd(), blueprint)
        if os.path.isfile(path) is False:
            raise ValueError
        with open(path) as f:
            return self.yaml.load(f)

    def parse(self, yaml):
        """
        Parses the content of a blueprint
        """
        for table in yaml:
            new_type = self.load_type(table)
            for field in yaml[table]:
                self.load_field(new_type, field)

    def load(self, filename):
        """
        Load a blueprint in the database
        """
        self.parse(self.read(filename))



# -*- coding: utf-8 -*-
from peewee import BooleanField, CharField, ForeignKeyField, IntegerField, SQL

from .Base import Base
from .Types import Types
from .Users import Users


class Fields(Base):
    name = CharField()
    field_type = CharField(default='string')
    unique = BooleanField(default=False, constraints=[SQL('DEFAULT false')])
    nullable = BooleanField(default=False, constraints=[SQL('DEFAULT false')])
    default_value = IntegerField(null=True)
    type_id = ForeignKeyField(Types)
    owner = ForeignKeyField(Users)



# -*- coding: utf-8 -*-
from peewee import BooleanField, CharField

from .Base import Base


class Users(Base):
    identifier = CharField(unique=True)
    superuser = BooleanField(default=False)

    def login(identifier):
        """
        Logs in an user, when provided with a valid identifer
        """
        try:
            return Users.get(Users.identifier == identifier)
        except Users.DoesNotExist:
            return None

    def _apply_permissions(self, query, model, level):
        """
        Applies the requested permission level to a query
        """
        lhs = (model.others_permission >= level)
        rhs_left_right = (model.owner_permission >= level)
        if model == Users:
            rhs_left = ((model.id == self.id) & rhs_left_right)
        else:
            rhs_left = ((model.owner == self.id) & rhs_left_right)
        rhs_right_right = (model.group_permission >= level)
        rhs_right = ((model.group == self.group) & rhs_right_right)
        return query.where(lhs | (rhs_left | rhs_right))

    def do(self, action, query, model):
        """
        Does an action with a query, on a model
        """
        actions = {
            'read': 1,
            'edit': 2,
            'eliminate': 3
        }
        return self._apply_permissions(query, model, actions[action])



# -*- coding: utf-8 -*-
import click

from peewee import OperationalError, ProgrammingError

from .App import App
from .Blueprints import Blueprints
from .Config import Config
from .Generator import Generator
from .Tokens import Tokens
from .models import Base, Fields, Types, Users, db


class Cli:

    @click.group()
    def main():
        pass

    @main.command()
    def quickstart():
        """
        Quickstart efesto, creating default tables
        """
        print('Setting up efesto...')
        config = Config()
        Base.init_db(config.db_url)
        try:
            db.create_tables([Fields, Types, Users])
        except OperationalError:
            print('An error occured during tables creation. '
                  'Please check your database credentials.')
            exit()
        except ProgrammingError:
            print('An error occurred during tables creation. '
                  'Please check your database.')
            exit()
        print('Quickstart finished successfully!')

    @main.command()
    @click.argument('user')
    @click.argument('expiration', default=100)
    def token(user, expiration):
        """
        Get an authentication token for an user
        """
        config = Config()
        token = Tokens.encode(config.jwt_secret, expiration=expiration,
                              sub=user, aud=config.jwt_audience)
        click.echo(token)

    @main.command()
    @click.argument('identifier')
    @click.option('--superuser', is_flag=True)
    def create_user(identifier, superuser):
        config = Config()
        Base.init_db(config.db_url)
        user = Users(identifier=identifier, group=1, owner_permission=1,
                     group_permission=1, others_permission=1,
                     superuser=superuser)
        user.save()

    @main.command()
    @click.argument('filename')
    def load_blueprint(filename):
        """
        Load the specified blueprint file
        """
        config = Config()
        Base.init_db(config.db_url)
        Blueprints().load(filename)

    @main.command()
    def generate():
        """
        Generates dynamic tables from existing types
        """
        config = Config()
        Base.init_db(config.db_url)
        types = Types.select().execute()
        generator = Generator()
        for dynamic_type in types:
            generator.generate(dynamic_type)
        db.create_tables(generator.models.values(), safe=True)

    @main.command()
    def run():
        return App.run()



# -*- coding: utf-8 -*-
from falcon import HTTPNotFound, HTTP_204

from peewee import DoesNotExist

import rapidjson

from ..Siren import Siren


class Items:
    def __init__(self, model):
        self.model = model

    def on_get(self, request, response, **params):
        """
        Executes a get request on a single item
        """
        user = params['user']
        query = self.model.select().where(self.model.id == params['id'])
        try:
            result = user.do('read', query, self.model).get()
        except DoesNotExist:
            raise HTTPNotFound()
        body = Siren(self.model, result, request.path)
        response.body = body.encode()

    def on_patch(self, request, response, **params):
        """
        Executes a patch request on a single item
        """
        user = params['user']
        query = self.model.select().where(self.model.id == params['id'])
        try:
            result = user.do('edit', query, self.model).get()
        except DoesNotExist:
            raise HTTPNotFound()
        json = rapidjson.load(request.bounded_stream)
        result.update_item(json)
        body = Siren(self.model, result, request.path)
        response.body = body.encode()

    def on_delete(self, request, response, **params):
        """
        Executes a delete request on a single item
        """
        user = params['user']
        query = self.model.delete().where(self.model.id == params['id'])
        user.do('eliminate', query, self.model).execute()
        response.status = HTTP_204



# -*- coding: utf-8 -*-
import os


class Config:

    defaults = {
        'db_url': 'postgres://postgres:postgres@localhost:5432/efesto',
        'jwt_secret': 'secret',
        'jwt_leeway': 5,
        'jwt_audience': 'efesto'
    }

    def __init__(self):
        self.apply()

    def apply(self):
        """
        Applies values, taking them from the environment or from the defaults
        """
        for key, value in self.defaults.items():
            setattr(self, key, os.getenv(key, default=value))

    def __getattribute__(self, name):
        """
        Gets an attribute or returns None
        """
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return None



# -*- coding: utf-8 -*-
from peewee import (BooleanField, CharField, DateTimeField, FloatField,
                    ForeignKeyField, IntegerField, SQL, TextField)

from .models import Base, Fields, Users


class Generator:
    """
    A model generator that is used to generated dynamically defined models.
    """

    mappings = {
        'string': CharField,
        'text': TextField,
        'int': IntegerField,
        'float': FloatField,
        'boolean': BooleanField,
        'date': DateTimeField
    }

    def __init__(self):
        self.models = {}

    def make_field(self, field, classname):
        """
        Generates a field from a field row
        """
        custom_field = CharField
        if field.field_type in self.mappings:
            custom_field = self.mappings[field.field_type]
        elif field.field_type in self.models:
            return ForeignKeyField(self.models[field.field_type],
                                   backref=classname)
        arguments = {'null': field.nullable, 'unique': field.unique}
        if field.default_value is not None:
            constraints = [SQL('DEFAULT {}'.format(field.default_value))]
            arguments['default'] = field.default_value
            arguments['constraints'] = constraints
        return custom_field(**arguments)

    def attributes(self, fields, classname):
        attributes = {}
        for field in fields:
            attributes[field.name] = self.make_field(field, classname)
        return attributes

    def new_model(self, type_instance):
        fields = Fields.select().where(Fields.type_id == type_instance.id)
        attributes = self.attributes(fields, type_instance.name)
        attributes['owner'] = ForeignKeyField(Users)
        model = type(type_instance.name, (Base, ), attributes)
        self.models[type_instance.name] = model

    def generate(self, type_instance):
        """
        Generate a model using a type
        """
        self.new_model(type_instance)
        return self.models[type_instance.name]



# -*- coding: utf-8 -*-
from efesto.Generator import Generator
from efesto.handlers import Collections, Items

import falcon


class Api:

    def __init__(self, **kwargs):
        self.api = falcon.API(**kwargs)

    @staticmethod
    def collection(model):
        return Collections(model)

    @staticmethod
    def item(model):
        return Items(model)

    def list_route(self, endpoint, model):
        self.api.add_route(endpoint, self.collection(model))

    def object_route(self, endpoint, model):
        route = '{}/{}'.format(endpoint, '{id}')
        self.api.add_route(route, self.item(model))

    def add_endpoint(self, endpoint, model):
        self.list_route(endpoint, model)
        self.object_route(endpoint, model)

    def dynamic_endpoints(self, types):
        generator = Generator()
        for dynamic_type in types:
            model = generator.generate(dynamic_type)
            self.add_endpoint('/{}'.format(dynamic_type.name), model)

    def cherries(self):
        """
        This method is the cherry on the cake
        """
        return self.api



# -*- coding: utf-8 -*
from falcon import HTTP_501

import rapidjson

from ..Siren import Siren


class Collections:
    def __init__(self, model):
        self.model = model

    def query(self, params):
        self.model.q = self.model.select()
        for key, value in params.items():
            self.model.query(key, value)
        return self.model.q

    def embeds(self, params):
        """
        Parses embeds and set joins on the query
        """
        embeds = params.pop('_embeds', None)
        if isinstance(embeds, str):
            embeds = [embeds]
        if embeds:
            for embed in embeds:
                property = getattr(self.model, embed)
                model = property.rel_model
                if hasattr(property, 'field'):
                    property = property.field
                    model = self.model
                self.model.q.join(model, on=(property == model.id))
            return embeds
        return []

    @staticmethod
    def page(params):
        return int(params.pop('page', 1))

    @staticmethod
    def items(params):
        return int(params.pop('items', 20))

    @staticmethod
    def apply_owner(user, payload):
        if 'owner_id' in payload:
            return None
        payload['owner_id'] = user.id

    def on_get(self, request, response, **params):
        """
        Executes a get request
        """
        user = params['user']
        page = self.page(request.params)
        items = self.items(request.params)
        query = self.query(request.params)
        embeds = self.embeds(request.params)
        result = user.do('read', query, self.model)
        paginated_query = result.paginate(page, items).execute()
        body = Siren(self.model, list(paginated_query), request.path,
                     page=page, total=result.count())
        response.body = body.encode(includes=embeds)

    def on_post(self, request, response, **params):
        json = rapidjson.load(request.bounded_stream)
        self.apply_owner(params['user'], json)
        item = self.model.create(**json)
        body = Siren(self.model, item, request.path)
        response.body = body.encode()

    def on_patch(self, request, response, **params):
        response.status = HTTP_501



# -*- coding: utf-8 -*-
import datetime

import jwt
from jwt.exceptions import DecodeError, ExpiredSignatureError


class Tokens:
    """
    Provides JWT encoding and decoding functionalities
    """

    @staticmethod
    def encode(secret, expiration=0, **kwargs):
        payload = {**kwargs}
        if expiration:
            now = datetime.datetime.utcnow()
            payload['exp'] = now + datetime.timedelta(100)
        return jwt.encode(payload, secret)

    @staticmethod
    def decode(secret, token):
        try:
            return jwt.decode(token, secret)
        except (DecodeError, ExpiredSignatureError):
            return None



# -*- coding: utf-8 -*-
from peewee import ModelSelect

import rapidjson


class Siren:

    def __init__(self, model=None, data=[], path='', total=0, page=1):
        self.model = model
        self.data = data
        self.path = path
        self.total = total
        self.page = page

    @staticmethod
    def paginate(path, items, current_page, total_items):
        links = [
            {'rel': ['self'], 'href': path}
        ]

        if total_items > len(items)*current_page:
            href = '{}?page={}'.format(path, current_page + 1)
            links.append({'rel': ['next'], 'href': href})

        if current_page != 1:
            href = '{}?page={}'.format(path, current_page - 1)
            links.append({'rel': ['previous'], 'href': href})
        return links

    @classmethod
    def _nested_entities(cls, include, item):
        """
        Parses nested entities. Used when a join is performed and thus
        there are nested item to encode.
        """
        nested = getattr(item, include)
        if isinstance(nested, ModelSelect):
            nested = list(nested)
            items = []
            path = '/{}'.format(nested[0].__class__.__name__)
            for item in nested:
                items.append(cls.entity(path, item))
            return items
        path = '/{}'.format(nested.__class__.__name__)
        return cls.entity(path, nested)

    @classmethod
    def entity(cls, path, item, includes=[]):
        """
        Creates an entity from a model instance
        """
        href = '{}/{}'.format(path, item.id)
        if path.endswith('/{}'.format(item.id)):
            href = path

        for include in includes:
            item.__data__[include] = cls._nested_entities(include, item)

        return {
            'properties': item.__data__,
            'class': [item.__class__.__name__],
            'links': [
                {'href': href, 'rel': 'self'}
            ]
        }

    def entities(self, includes=[]):
        entities = []
        for item in self.data:
            entities.append(self.entity(self.path, item, includes=includes))

        fields = []
        name = 'add-item'
        if self.model:
            fields = self.model.get_columns()
            name = 'add-' + self.model._meta.name

        actions = [
            {'name': name, 'method': 'POST', 'type': 'application/json',
             'fields': fields}
        ]
        links = self.paginate(self.path, self.data, self.page, self.total)
        return {'entities': entities, 'actions': actions, 'links': links}

    def encode(self, *args, includes=[]):
        if isinstance(self.data, list):
            return rapidjson.dumps(self.entities(includes=includes),
                                   datetime_mode=1)
        return rapidjson.dumps(self.entity(self.path, self.data),
                               datetime_mode=1)



# -*- coding: utf-8 -*-
from efesto.models import Users

from falcon import HTTPUnauthorized

import jwt
from jwt.exceptions import DecodeError, ExpiredSignatureError


class Authentication:

    def __init__(self, secret, audience):
        self.secret = secret
        self.audience = audience

    @staticmethod
    def unauthorized():
        """
        Raises a 401 error
        """
        raise HTTPUnauthorized('Login required', 'Please login',
                               ['Bearer realm="login required"'])

    def bearer_token(self, auth_header):
        """
        Get the token from the auth header
        """
        shards = auth_header.split()
        if len(shards) == 2:
            if shards[0] == 'Bearer':
                return shards[1]
        return self.unauthorized()

    def decode(self, token):
        """
        Decode a token
        """
        try:
            return jwt.decode(token, self.secret, audience=self.audience)
        except (DecodeError, ExpiredSignatureError):
            return self.unauthorized()

    def login(self, payload):
        if 'sub' in payload:
            return Users.login(payload['sub'])
        return self.unauthorized()

    def process_resource(self, request, response, resource, params):
        if request.auth is None:
            return self.unauthorized()

        token = self.bearer_token(request.auth)
        payload = self.decode(token)
        user = self.login(payload)
        if user is None:
            return self.unauthorized()
        params['user'] = user



#!/usr/bin/env python
import io
import os
import sys

from efesto.Version import version

from setuptools import find_packages, setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()


readme = io.open('README.md', 'r', encoding='utf-8').read()

setup(
    name='efesto',
    description='RESTful (micro)server that can build an API in minutes.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/getefesto/efesto',
    author='Jacopo Cascioli',
    author_email='hi@jacopocascioli.com',
    license='GPL3',
    version=version,
    packages=find_packages(),
    tests_require=[
        'pytest',
        'pytest-mock',
        'pytest-falcon'
    ],
    setup_requires=['pytest-runner'],
    install_requires=[
        'falcon>=1.1.0',
        'falcon-cors>=1.0.1',
        'psycopg2-binary>=2.7.4',
        'peewee>=3.2.0',
        'click>=6.7',
        'colorama>=0.3.3',
        'aratrum>=0.3.2',
        'python-rapidjson>=0.6.2',
        'pyjwt>=1.5.2',
        'ruamel.yaml>=0.15.37'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
    entry_points="""
        [console_scripts]
        efesto=efesto.Cli:Cli.main
    """
)



# -*- coding: utf-8 -*-
from peewee import IntegerField, Model, SQL

from playhouse import db_url

from .Database import db


class Base(Model):

    conversions = {
        'AutoField': 'number',
        'BooleanField': 'number',
        'CharField': 'text',
        'TextField': 'text',
        'DateTimeField': 'date',
        'FloatField': 'number',
        'ForeignKeyField': 'number',
        'IntegerField': 'number',
        'PrimaryKeyField': 'number'
    }

    class Meta:
        database = db

    @classmethod
    def get_columns(cls):
        """
        Produces a list of columns for the current model.
        """
        columns = []
        for name, column in cls._meta.fields.items():
            column_type = cls.conversions[column.__class__.__name__]
            columns.append({'name': name, 'type': column_type})
        return columns

    @staticmethod
    def init_db(url):
        dictionary = db_url.parse(url)
        name = dictionary.pop('database')
        db.init(name, **dictionary)

    def update_item(self, data):
        for key, value in data.items():
            setattr(self, key, value)
        self.save()

    @classmethod
    def filter(cls, key, value, operator):
        """
        Adds a filter to the current query
        """
        if operator == '!':
            return cls.q.where(getattr(cls, key) != value)
        elif operator == '>':
            return cls.q.where(getattr(cls, key) > value)
        elif operator == '<':
            return cls.q.where(getattr(cls, key) < value)
        return cls.q.where(getattr(cls, key) == value)

    @classmethod
    def query(cls, key, value):
        """
        Builds a select query
        """
        if hasattr(cls, key) is False:
            return None
        operator = None
        if value[0] in ['!', '>', '<']:
            operator = value[0]
            value = value[1:]
        cls.q = cls.filter(key, value, operator)

    group = IntegerField(default=1, constraints=[SQL('DEFAULT 1')])
    owner_permission = IntegerField(default=3, constraints=[SQL('DEFAULT 3')])
    group_permission = IntegerField(default=1, constraints=[SQL('DEFAULT 1')])
    others_permission = IntegerField(default=0, constraints=[SQL('DEFAULT 0')])



# -*- coding: utf-8 -*-
from .Api import Api
from .Config import Config
from .Middlewares import Authentication
from .models import Base, Fields, Types, Users


class App:

    def run():
        """
        Runs efesto
        """
        config = Config()
        Base.init_db(config.db_url)
        middleware = Authentication(config.jwt_secret, config.jwt_audience)
        api = Api(middleware=middleware)
        api.add_endpoint('/users', Users)
        api.add_endpoint('/fields', Fields)
        api.add_endpoint('/types', Types)
        types = Types.select().execute()
        api.dynamic_endpoints(types)
        return api.cherries()
